/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 950 $ $LastChangedDate: 2002-09-12 23:10:49 +0200 (Thu, 12 Sep 2002) $
 *
 * Runtime environment for gont.
 */

#include <gont/runtime.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

struct gont_string {
	gont_int len;
	unsigned char data[0];
};

static struct gont_string gont_empty_string_data = {0};
struct gont_string *gont_empty_string = &gont_empty_string_data;

#define ONE(c, ptr, check, left)		\
	c <<= 6;				\
	if (check && (*ptr & 0xc0) != 0x80) 	\
		goto bad_utf;			\
	if (check && --left < 0)		\
		goto short_utf;			\
	c |= *ptr++ & 0x3f;
		
#define GET_UTF8(c, ptr, check, left) 		\
do {						\
	if (check && --left < 0)		\
		goto short_utf;			\
	if ((*ptr & 0x80) == 0) {		\
		c = *ptr++;			\
	} else if ((*ptr & 0xe0) == 0xc0) {	\
		c = *ptr++ & 0x1f;		\
		ONE(c, ptr, check, left);	\
	} else if ((*ptr & 0xf0) == 0xe0) {	\
		c = *ptr++ & 0x0f;		\
		ONE(c, ptr, check, left);	\
		ONE(c, ptr, check, left);	\
	} else if ((*ptr & 0xf8) == 0xf0) {	\
		c = *ptr++ & 0x07;		\
		ONE(c, ptr, check, left);	\
		ONE(c, ptr, check, left);	\
		ONE(c, ptr, check, left);	\
	} else if ((*ptr & 0xfc) == 0xf8) {	\
		c = *ptr++ & 0x03;		\
		ONE(c, ptr, check, left);	\
		ONE(c, ptr, check, left);	\
		ONE(c, ptr, check, left);	\
		ONE(c, ptr, check, left);	\
	} else if ((*ptr & 0xfe) == 0xfc) {	\
		c = *ptr++ & 0x01;		\
		ONE(c, ptr, check, left);	\
		ONE(c, ptr, check, left);	\
		ONE(c, ptr, check, left);	\
		ONE(c, ptr, check, left);	\
		ONE(c, ptr, check, left);	\
	} else					\
		goto bad_utf;			\
} while (0)

#define SKIP(ptr, n) 				\
do {						\
	gont_int i;				\
	gont_int dummy;				\
	int dummy2;				\
	for (i = 0; i < n; i++)			\
		GET_UTF8(dummy, ptr, 0, dummy2);\
} while (0)


#define BITS(k) ((1ul << (k)) - 1ul)
enum {
	mask_1 = BITS(7 + 6 * 0),
	mask_2 = BITS(5 + 6 * 1),
	mask_3 = BITS(4 + 6 * 2),
	mask_4 = BITS(3 + 6 * 3),
	mask_5 = BITS(2 + 6 * 4),
	mask_6 = BITS(1 + 6 * 5)
};

#define UTF8_ADDLEN(len, c)				\
do {							\
	if ((c & ~mask_1) == 0) len += 1;		\
	else if ((c & ~mask_2) == 0) len += 2;		\
	else if ((c & ~mask_3) == 0) len += 3;		\
	else if ((c & ~mask_4) == 0) len += 4;		\
	else if ((c & ~mask_5) == 0) len += 5;		\
	else if ((c & ~mask_6) == 0) len += 6;		\
	else goto invalid_unicode;			\
} while (0)

#define STORE(c, ptr, n) \
	*ptr++ = ((c >> (6 * n)) & 0x3f) | 0x80
	
#define STORE_UTF8(c, ptr)				\
do {							\
	if ((c & ~mask_1) == 0) {			\
		*ptr++ = c;				\
	} else if ((c & ~mask_2) == 0) {		\
		*ptr++ = (c >> 6 * 1) | 0xc0;		\
		STORE(c, ptr, 0);			\
	} else if ((c & ~mask_3) == 0) {		\
		*ptr++ = (c >> 6 * 2) | 0xe0;		\
		STORE(c, ptr, 1);			\
		STORE(c, ptr, 0);			\
	} else if ((c & ~mask_4) == 0) {		\
		*ptr++ = (c >> 6 * 3) | 0xf0;		\
		STORE(c, ptr, 2);			\
		STORE(c, ptr, 1);			\
		STORE(c, ptr, 0);			\
	} else if ((c & ~mask_5) == 0) {		\
		*ptr++ = (c >> 6 * 4) | 0xf8;		\
		STORE(c, ptr, 3);			\
		STORE(c, ptr, 2);			\
		STORE(c, ptr, 1);			\
		STORE(c, ptr, 0);			\
	} else if ((c & ~mask_6) == 0) {		\
		*ptr++ = (c >> 6 * 5) | 0xfc;		\
		STORE(c, ptr, 4);			\
		STORE(c, ptr, 3);			\
		STORE(c, ptr, 2);			\
		STORE(c, ptr, 1);			\
		STORE(c, ptr, 0);			\
	} else						\
		goto invalid_unicode;			\
} while (0)

static struct gont_string *make(gont_int len)
{
	struct gont_string *s;

	s = gont_xmalloc(sizeof(*s) + len);
	s->len = -1;

	return s;
}

static gont_int count_bytes(unsigned char *utf, gont_int chars)
{
	unsigned char *ptr = utf;

	SKIP(ptr, chars);
	
	return ptr - utf;

bad_utf:
short_utf:
	abort();
}

static gont_int string_bytes(struct gont_string *s)
{
	return count_bytes(s->data, s->len);
}

unsigned char *gont_string_to_utf8(unsigned char *buf_hint, gont_int buf_size,
				   struct gont_string *s, gont_int *sizep)
{
	unsigned char *ret;
	gont_int size;
	
	size = string_bytes(s);
	if (size < buf_size)
		ret = buf_hint;
	else
		ret = gont_xmalloc_atomic(size + 1);
	memcpy(ret, s->data, size);
	ret[size] = '\0';

	if (sizep)
		*sizep = size;
	
	return ret;
}

gont_int *gont_string_to_utf32(struct gont_string *s)
{
	gont_int *ret, n, i;
	int dummy = 0;
	unsigned char *ptr;

	n = s->len;
	if (n == 0)
		return gont_xmalloc_atomic(sizeof(gont_int));

	ptr = s->data;
	ret = gont_xmalloc_atomic(sizeof(gont_int) * n);

	for (i = 0; i < n; i++)
		GET_UTF8(ret[i], ptr, 0, dummy);

	return ret;
	
bad_utf:
short_utf:
	abort();
}

void gont_string_iter(struct gont_lambda_expr *f, struct gont_string *s)
{
	gont_int ch, n, i;
	int dummy = 0;
	unsigned char *ptr;

	n = s->len;
	ptr = s->data;

	for (i = 0; i < n; i++) {
		GET_UTF8(ch, ptr, 0, dummy);
		f->fnc(f->closure, ch);
	}

	return;
	
bad_utf:
short_utf:
	abort();
}

struct gont_string *gont_string_of_utf8(const unsigned char *ptr, gont_int bytes)
{
	struct gont_string *s;
	unsigned char *dest;
	long left = bytes;
	gont_int c, len = 0;

	if (bytes == 0)
		return gont_empty_string;

	/* This will be enough, but we might allocate a bit too much if
	   input string uses non-optimal encoding.  */
	s = make(bytes);
	dest = s->data;

	while (left) {
		GET_UTF8(c, ptr, 1, left);
		STORE_UTF8(c, dest);
		len++;
	}

	s->len = len;

	return s;

bad_utf:
	gont_invalid_mbseq();
	
short_utf:
	gont_incomplete_mbseq();

invalid_unicode:
	/* Not possible if encoding from utf8.  */
	abort();
}

struct gont_string *gont_string_of_utf32(const gont_int *ptr, gont_int chars)
{
	struct gont_string *s;
	unsigned char *dest;
	gont_int len, i;

	if (chars == 0)
		return gont_empty_string;

	len = 0;

	for (i = 0; i < chars; i++)
		UTF8_ADDLEN(len, ptr[i]);
	
	/* This will be enough, but we might allocate a bit too much if
	   input string uses non-optimal encoding.  */
	s = make(len);
	dest = s->data;

	len = 0;

	for (i = 0; i < chars; i++) {
		STORE_UTF8(ptr[i], dest);
		len++;
	}

	s->len = len;

	return s;

invalid_unicode:
	gont_invalid_unicode_character();
}

struct gont_string *gont_strjoin(struct gont_string *s1, 
			         struct gont_string *s2)
{
	gont_int len1, len2;
	struct gont_string *r;
	
	len1 = string_bytes(s1);
	len2 = string_bytes(s2);
	
	r = make(len1 + len2);

	memcpy(r->data, s1->data, len1);
	memcpy(r->data + len1, s2->data, len2);

	r->len = s1->len + s2->len;

	return r;
}

gont_int gont_streq(struct gont_string *s1, struct gont_string *s2)
{
	gont_int len1, len2;
	
	if (s1->len != s2->len)
		return 0;
	
	len1 = string_bytes(s1);
	len2 = string_bytes(s2);

	if (len1 != len2)
		return 0;
	
	return memcmp(s1->data, s2->data, len1) == 0;
}

gont_int gont_strcmp(struct gont_string *s1, struct gont_string *s2)
{
	gont_int len1, len2;
	int r;
	
	len1 = string_bytes(s1);
	len2 = string_bytes(s2);
	
	if (len1 < len2) {
		r = memcmp(s1->data, s2->data, len1);
		return r == 0 ? -1 : r;
	} else if (len1 > len2) {
		r = memcmp(s1->data, s2->data, len2);
		return r == 0 ? 1 : r;
	} else
		return memcmp(s1->data, s2->data, len1);
}

gont_int gont_string_hash(struct gont_string *s)
{
	gont_int pos = 0, hash = 0, len;

	len = string_bytes(s);
	
	while (pos < len) {
		hash += s->data[pos++];
		hash *= 37;
	}

	return hash & 0x7fffffff;
}

gont_int gont_string_length(struct gont_string *s)
{
	return s->len;
}

struct gont_string *gont_string_sub(struct gont_string *s, gont_int pos, 
				    gont_int len)
{
	unsigned char *beg;
	unsigned char *ptr;
	struct gont_string *r;
	
	/* Note that first check isn't redundant (overflow). */
	if (pos > s->len || pos + len > s->len)
		gont_invalid_argument("gont_strsub");
	
	if (len == 0)
		return gont_empty_string;
	
	ptr = s->data;

	SKIP(ptr, pos);
	beg = ptr;
	SKIP(ptr, len);

	r = make(ptr - beg);
	memcpy(r->data, beg, ptr - beg);
	r->len = len;
		
	return r;

bad_utf:
short_utf:
	abort();
}

gont_int gont_string_get(struct gont_string *s, gont_int idx)
{
	gont_int r = 0;
	int dummy = 0;
	unsigned char *ptr = s->data;
	
	if (idx >= s->len)
		gont_invalid_argument("gont_string_get");

	idx++;
	
	while (idx--)
		GET_UTF8(r, ptr, 0, dummy);
		
	return r;

bad_utf:
short_utf:
	abort();
}

struct gont_string *gont_string_make(gont_int n, gont_int c)
{
	unsigned char buf[10], *ptr = buf;
	struct gont_string *s;
	
	STORE_UTF8(c, ptr);

	s = make((ptr - buf) * n);
	s->len = n;

	if (ptr - buf == 1) {
		memset(s->data, *buf, n);
	} else {
		int len = ptr - buf;
		ptr = s->data;
		while (n--) {
			memcpy(ptr, buf, len);
			ptr += len;
		}
	}

	return s;

invalid_unicode:
	gont_invalid_unicode_character();
}

gont_int gont_string_index_from(struct gont_string *s, gont_int pos, gont_int c)
{
	int dummy;
	unsigned char *ptr;
	gont_int c2;

	if (pos >= s->len)
		return -2;
	
	ptr = s->data;
	SKIP(ptr, pos);
	
	while (pos < s->len) {
		GET_UTF8(c2, ptr, 0, dummy);
		if (c2 == c)
			return pos;
		pos++;
	}

	return -1;

bad_utf:
short_utf:
	abort();
}	

gont_int gont_string_rindex_from(struct gont_string *s, gont_int pos, gont_int c)
{
	int dummy;
	unsigned char *ptr;
	gont_int c2, i, last = -1;

	if (pos >= s->len)
		return -2;
	
	ptr = s->data;
	
	for (i = 0; i <= pos; i++) {
		GET_UTF8(c2, ptr, 0, dummy);
		if (c2 == c)
			last = i;
	}

	return last;

bad_utf:
short_utf:
	abort();
}	

gont_int gont_string_index_str_from(struct gont_string *s, gont_int pos, 
				    struct gont_string *c)
{
	unsigned char *ptr;
	int dummy;
	gont_int dummy2, len, max;
	
	if (pos > s->len)
		return -2;
		
	if (c->len == 0)
		return pos;
	
	if (c->len > s->len)
		return -1;
	
	ptr = s->data;
	SKIP(ptr, pos);

	len = string_bytes(c);
	
	for (max = s->len - c->len; pos <= max; pos++) {
		if (memcmp(c->data, ptr, len) == 0)
			return pos;
		GET_UTF8(dummy2, ptr, 0, dummy);
	}

	return -1;

bad_utf:
short_utf:
	abort();
}

int gont_append_approx(struct gont_lambda_expr *approx_fnc, 
		       char **in_buf, size_t *in_left,
		       char **out_buf, size_t *out_left)
{
	char buf[20], *approx;
	unsigned char *ptr;
	gont_int ch, approx_len;
	int dummy = 0;

	ptr = *in_buf;
	GET_UTF8(ch, ptr, 0, dummy);

	if (approx_fnc) {
		struct gont_string *s = 
			approx_fnc->fnc(approx_fnc->closure, ch);

		/* We require approx_fnc to return just ASCII.  */
		approx = s->data;
		approx_len = s->len;
	} else {
		sprintf(buf, "\\x{%lx}", ch);
		approx = buf;
		approx_len = strlen(approx);
	}

	if (*out_left < approx_len) {
		errno = E2BIG;
		return -1;
	} else {
		memcpy(*out_buf, approx, approx_len);
		*out_buf += approx_len;
		*out_left -= approx_len;
		*in_left -= (char*)ptr - *in_buf;
		*in_buf = ptr;
		errno = 0;
		return approx_len;
	}

bad_utf:
short_utf:
	abort();
}
